# -*- coding: utf-8 -*-
"""
Created on Mon Jun  7 09:34:06 2021

@author: Frédéric Mandon
www.maths-info-lycee.fr
"""

import tkinter as tk
# import tkMessageBox
from classe_machine import *
from classe_gui_saisie import *
from random import randint

class GUITuringExecution:
    
    def __init__(self):
        self.fenetre = tk.Tk()
        #  VARIABLES METIER récupérées pour l'animation
        self.machine = MachineTuring()
        # VARIABLES POUR L'INTERFACE GRAPHIQUE
        self.vitesse = 1            #  vitesse animations (en ticks)
        self.texte_symboles = {0: "0", 1 : "1", 2 : "\u25a2"}
        self.texte_dep = {-1 : "déplacement vers la gauche", 
                          1 : "déplacement vers la droite", 
                          0 : "tête immobile"}
        self.lg_ruban_ig = 24       #  Obligatoire, compléter avec des 2 le ruban de
                                    #  longueur 20 dans classe_machine si nécessaire
        self.nb_etats_ig = 10       #  Obligatoire, il peut y avoir des états "vides"
        # Tailles des différents éléments par défaut :
        self.largeur_cell_etat = 50
        self.largeur_cell_ruban = 25
        self.taille_police = 8
        # Etiquette de la transition en cours et du résumé de la machine
        self.label_transition = None
        self.label_resume = None
        self.deroulement = None
       
        #  INITIALISATION FENETRE   
        #  parent_window.geometry("width_size x height_size + x_position + y_position")
        #  Calcul des tailles des cellules de la grille, puis de la fenêtre et du canevas
        #  Tailles entre 15/30 et 51/102 respectivement

        self.fenetre.title('Machine de Turing')
        self.fenetre.configure(background='white')
        largeur_ecran = self.fenetre.winfo_screenwidth()
        hauteur_ecran = self.fenetre.winfo_screenheight()
        self.largeur_cell_ruban = max(min(largeur_ecran // self.lg_ruban_ig, 51), 15)
        if self.largeur_cell_ruban % 3 != 0:    #  pour des compatibiltés de taille
            self.largeur_cell_ruban += 1        #  avec la police divisible par 3
            if self.largeur_cell_ruban % 3 != 0:
                self.largeur_cell_ruban += 1
        self.largeur_cell_etat = 2 * self.largeur_cell_ruban
        self.taille_police = self.largeur_cell_ruban//3
        for i in range(self.lg_ruban_ig) :
            self.fenetre.columnconfigure(i, minsize = self.largeur_cell_ruban)
        self.fenetre.rowconfigure(0,minsize = self.largeur_cell_ruban)
        self.fenetre.rowconfigure(2,minsize = self.largeur_cell_ruban)
        
        # Canevas
        h_canevas = hauteur_ecran // 2.5
        l_canevas = self.largeur_cell_ruban*(self.lg_ruban_ig)
        self.can = tk.Canvas(self.fenetre, bg='white', width = l_canevas, 
                 height = h_canevas, highlightthickness=0)
        self.can.grid(row = 1, column = 0, columnspan = self.lg_ruban_ig)

        #  Ruban dans le canevas
        x_r_0 = 0
        y_r_0 = h_canevas
        y_r_1 = h_canevas * 0.9
        self.ruban_can = []
        for i in range(self.lg_ruban_ig):
            self.ruban_can.append(self.can.create_rectangle(
                x_r_0, y_r_0, x_r_0 + self.largeur_cell_ruban, y_r_1,
                outline = 'black', fill = 'firebrick1'))
            x_r_0 = x_r_0 + self.largeur_cell_ruban
        self.can.itemconfig(self.ruban_can[2], fill='firebrick3')

        x_r_0 = self.largeur_cell_ruban // 2
        y_r_0 = h_canevas * 0.95
        ruban_alea = [self.texte_symboles[randint(0,4)//2] for i in range(self.lg_ruban_ig)]
        self.ruban_texte_can = []
        for i in range(self.lg_ruban_ig):
            self.ruban_texte_can.append(
                self.can.create_text(x_r_0, y_r_0, text= ruban_alea[i],
                                 fill="black",font=('Arial', self.taille_police)))
            x_r_0 = x_r_0 + self.largeur_cell_ruban
        self.can.itemconfig(self.ruban_texte_can[2], fill='white')
            
        #  Etats dans le canevas
        x_r_0 = self.largeur_cell_etat
        y_r_0 = 0
        y_r_1 = h_canevas * 0.1
        self.etats_can = []
        self.etat_0_canevas = 10
        for i in range(self.nb_etats_ig):
            self.etats_can.append(self.can.create_rectangle(
                x_r_0, y_r_0, x_r_0 + self.largeur_cell_etat, y_r_1,
                outline = 'black', fill = 'turquoise3'))
            x_r_0 = x_r_0 + self.largeur_cell_etat
        self.can.itemconfig(self.etats_can[0], fill='turquoise4')  
            
        x_r_0 = self.largeur_cell_etat * 1.5
        y_r_0 = h_canevas * 0.05
        self.etats_texte_can = []
        texte = ['INIT'] +\
            ["ETAT " + str(i) for i in range(1,self.nb_etats_ig - 1)] +\
                ['FIN']
        for i in range(self.nb_etats_ig):
            self.etats_texte_can.append(
                self.can.create_text(x_r_0, y_r_0, text= texte[i],
                                 fill="black",font=('Arial', self.taille_police)))
            x_r_0 = x_r_0 + self.largeur_cell_etat
        self.can.itemconfig(self.etats_texte_can[0], fill='white')
        
        #  coordonnées tête et lien vers état
        self.x0_t = self.largeur_cell_ruban*2.5     #  _t pour triangle _d pour départ
        self.y0_t = h_canevas * 0.9
        self.x1_t = self.largeur_cell_ruban*3.5
        self.y1_t = int(h_canevas * 0.56)
        self.x2_t = self.largeur_cell_ruban*1.5
        self.y2_t = int(h_canevas * 0.56)
        self.x0_s = self.largeur_cell_ruban*2.5      #  _s pour segment _d pour départ
        self.y0_s = int(h_canevas * 0.56)
        self.x1_s = self.largeur_cell_etat * 1.5
        self.y1_s = h_canevas * 0.1
        self.tete_lecture = self.can.create_polygon(self.x0_t, self.y0_t, self.x1_t, 
                                          self.y1_t, self.x2_t, self.y2_t, 
                                           fill = 'OliveDrab3',
                                           outline = 'OliveDrab4', width = 2)
        self.lien_etat = self.can.create_line(self.x0_s, self.y0_s, self.x1_s, self.y1_s, 
                                fill = 'OliveDrab4', width = 3)
        
         #  Fonctionnement de la machine
        self.label_resume =tk.Label(self.fenetre, text = "Pas de machine chargée", 
                                    font=("Arial", self.taille_police),
                                      borderwidth=1, relief="raised", padx = 2, pady = 2,
                                      bg = "ivory1")
        self.label_transition = tk.Label(self.fenetre, text = "Transition", 
                                         font=("Arial", self.taille_police - 2),
                                      borderwidth=0, relief="solid", padx = 2, pady = 2)
        self.bouton_test1 = tk.Button(self.fenetre, text="Exemple 1", 
                                      command = lambda : self.test(1))
        self.bouton_test2 = tk.Button(self.fenetre, text="Exemple 2", 
                                      command = lambda : self.test(2))
        self.bouton_saisie = tk.Button(self.fenetre, text="Saisir une machine", 
                                      command = lambda : self.saisie(False))
        self.bouton_modifier = tk.Button(self.fenetre, text="Modifier la machine", 
                                      command = lambda : self.saisie(True))
        self.bouton_mouvement = tk.Button(self.fenetre, text="On tourne !", 
                                          command = lambda : self.mouvement())
        self.bouton_quitter = tk.Button(self.fenetre, text="Quitter", command = self.fenetre.destroy)
   
    def chaineTransition(self, ind_etape):
        """
        routine d'écriture du label des transitions lors d'une etape du déroulement
        Une étape est un tuple:
            (ruban, indice position tête sur le ruban,
             symbole_lu, état courant, symbole écrit,
             état suivant, déplacement)
        """
        etape = self.machine.deroulement[ind_etape]
        lu = etape[2]
        etat_courant = etape[3]
        ecrit = etape[4]
        etat_suivant = etape[5]
        dep = etape[6]
        t_lu = self.texte_symboles[lu ]
        t_ecrit = self.texte_symboles[ecrit]
        t_dep = self.texte_dep[dep]
        t_etat_courant = self.machine.etats[etat_courant]
        t_etat_suivant = self.machine.etats[etat_suivant]
        texte_trans = "Symbole lu : " + t_lu + \
            " , Etat courant : " + t_etat_courant + "  \u2192  " + \
                "Symbole écrit : " + t_ecrit + \
            " , Etat suivant : " + t_etat_suivant +" , " + t_dep
        return texte_trans
    
    def initGUI(self):
        self.label_resume.grid(row=4, column = 0, columnspan = self.lg_ruban_ig)
        self.label_transition.grid(row=5, column = 0, columnspan = self.lg_ruban_ig)
        self.bouton_test1.grid(row=6, column = 0, columnspan = 3, pady = 5)
        self.bouton_test2.grid(row=6, column = 3, columnspan = 3, pady = 5)
        self.bouton_saisie.grid(row=6, column = 6, columnspan = 3, pady = 5)
        self.bouton_modifier.grid(row=6, column = 9, columnspan = 3, pady = 5)
        self.bouton_mouvement.grid(row=6, column = 12, columnspan = 3, pady = 5)
        self.bouton_quitter.grid(row=6, column = 20, pady = 5)
        self.fenetre.mainloop()
    
    def saisie(self, modif):
        if not modif :
            self.machine = MachineTuring()
        etats_standarts = (self.machine.ETAT_INIT, self.machine.CODE_INIT, 
                           self.machine.ETAT_FIN, self.machine.CODE_FIN,
                           self.machine.ETAT_VRAI, self.machine.CODE_VRAI, 
                           self.machine.ETAT_FAUX, self.machine.CODE_FAUX)
        nb_etats = self.machine.NB_ETATS
        lg_ruban = self.machine.LONGUEUR_RUBAN
        fen_fille = tk.Toplevel()
        saisie = GUISaisieTuring(fen_fille, self.taille_police, etats_standarts, lg_ruban, nb_etats,
                                 self.machine, modif, self)
        saisie.initGUI(self.machine)

        
        
    def reInitGui(self) :
        """
        rénitialisation de la fenêtre sur une machine rentrée en mémoire
        """        
        #  Réinitialisation du ruban
        ruban_der = [2, 2] + self.machine.deroulement[0][0] + [2, 2]
        for i in range(len(ruban_der)):
            self.can.itemconfig(self.ruban_can[i], fill = 'firebrick1')
            self.can.itemconfig(self.ruban_texte_can[i], fill = 'black',
                                text = self.texte_symboles[ruban_der[i]])
        ind_debut_ruban = 1
        while ruban_der[ind_debut_ruban + 1] == 2:
            ind_debut_ruban = ind_debut_ruban + 1
        self.can.itemconfig(self.ruban_can[ind_debut_ruban], fill = 'firebrick3')
        self.can.itemconfig(self.ruban_texte_can[ind_debut_ruban], fill = 'white')

        #  Etats dans le canevas, on les centre dans le canevas
        nb_etats = len(list(self.machine.etats.keys()))
        self.etat_0_canevas = (self.nb_etats_ig - nb_etats) // 2

        for i in range(self.etat_0_canevas):
            # remplissage éventuel de blancs à gauche des états et fin sur "index de état de démmarage -1"
            self.can.itemconfig(self.etats_can[i], outline = 'white', fill='white')
            self.can.itemconfig(self.etats_texte_can[i], text = "",fill='white')
        if self.etat_0_canevas == 0:
            # sinon il faut initialiser "index de état de démmarage -1"
            i = -1
        for etat in self.machine.etats:
            i = i + 1
            if etat == 0 :      # à affiner en turquoise4 pour changer la couleur des états en cours
                self.can.itemconfig(self.etats_can[i], fill='turquoise4', outline = 'black')
                self.can.itemconfig(self.etats_texte_can[i], 
                                    text = self.machine.etats[etat], fill='white')
                ind_debut_etat = i
            else :
                self.can.itemconfig(self.etats_can[i], fill='turquoise3', outline = 'black')
                self.can.itemconfig(self.etats_texte_can[i], 
                                    text = self.machine.etats[etat], fill='black')
        i = i + 1
        for j in range(i, self.nb_etats_ig):
            self.can.itemconfig(self.etats_can[j], outline = 'white', fill='white')
            self.can.itemconfig(self.etats_texte_can[j], text = "", fill='white')
        #  l'état le plus à droite est écrit après les états vides pour la bordure noire
        #  mais ça marche pas quand même lol :
        self.can.itemconfig(self.etats_can[i - 1], fill='turquoise3', outline = 'black')
            
        #  réinitialisation tete et lien
        self.x0_t = self.largeur_cell_ruban*(ind_debut_ruban + 0.5)     #  _t pour triangle _d pour départ
        self.x1_t = self.largeur_cell_ruban*(ind_debut_ruban - 0.5)
        self.x2_t = self.largeur_cell_ruban*(ind_debut_ruban + 1.5)
        self.x0_s = self.largeur_cell_ruban*(ind_debut_ruban + 0.5)      #  _s pour segment _d pour départ
        self.x1_s = self.largeur_cell_etat*(ind_debut_etat + 1.5)
        self.can.delete(self.lien_etat)
        self.can.delete(self.tete_lecture)
        self.tete_lecture = self.can.create_polygon(self.x0_t, self.y0_t, self.x1_t, 
                                      self.y1_t, self.x2_t, self.y2_t, 
                                       fill = 'OliveDrab3',
                                       outline = 'OliveDrab4', width = 2)
        self.lien_etat = self.can.create_line(self.x0_s, self.y0_s, self.x1_s, self.y1_s, 
                                fill = 'OliveDrab4', width = 3)
        self.can.update()
        
        #  Réinitialisation des étiquettes résumé et transition
        self.label_resume.configure(text = self.machine.resume)  
        self.label_resume.update()
        texte_transition =  self.chaineTransition(0)
        self.label_transition.configure(text = texte_transition)
        self.label_transition.update()    
        return
    
    def test(self, num_test):
        if num_test == 1:
            self.machine.exemple1()
        else :
            self.machine.exemple2()

        ok = self.machine.turing()
        if not ok:
            assert "Vous foirâtes votre machine"
        else:
            #for ligne in machine.deroulement :
            #    print(ligne)
            self.deroulement = self.machine.deroulement
            self.etats = self.machine.etats

        self.reInitGui()

    
    def mvtTeteSeule(self, dep, ind_ruban, ecrit):
        #  l'état ne change pas, seul le ruban change
        self.can.itemconfig(self.ruban_can[ind_ruban], fill = 'chocolate1')
        self.can.itemconfig(self.ruban_texte_can[ind_ruban], fill = 'black',
                                text = self.texte_symboles[ecrit])
        self.can.update()
        self.can.after(500)
        self.can.itemconfig(self.ruban_can[ind_ruban], fill = 'firebrick1')
        for i in range(self.largeur_cell_ruban):
            self.can.move(self.tete_lecture, dep, 0)
            self.x0_t = self.x0_t + dep
            self.can.delete(self.lien_etat)
            self.lien_etat = self.can.create_line(self.x0_t, self.y0_s, self.x1_s, self.y1_s, 
                                fill = 'OliveDrab4', width = 3)
            self.can.update()
            self.can.after(self.vitesse)
        self.can.itemconfig(self.ruban_can[ind_ruban + dep], fill = 'firebrick3')
        self.can.itemconfig(self.ruban_texte_can[ind_ruban + dep], fill = 'white')
        self.can.update()
        
    def mvtTeteEtat(self, dep, ind_ruban, ecrit, etat_courant, etat_suivant):
        self.can.itemconfig(self.ruban_can[ind_ruban], fill = 'chocolate1')
        self.can.itemconfig(self.ruban_texte_can[ind_ruban], fill = 'black',
                                text = self.texte_symboles[ecrit])
        self.can.update()
        self.can.after(500)
        self.can.itemconfig(self.ruban_can[ind_ruban], fill = 'firebrick1')
        # les dictionnaires étant ordonnés maintenant, on peut allègrement utiliser
        # un pseudo indice pour trouver l'état à atteindre et calculer ainsi
        # le déplacement côté états. Pour plus propre, il faudrait, lors de 
        # la réinitialisation de la GUI, créer une liste avec les états rangés
        # par indice
        liste_etats = list(self.machine.etats.keys())
        i_etat_c = liste_etats.index(etat_courant)
        i_etat_s = liste_etats.index(etat_suivant)
        
        self.can.itemconfig(self.etats_can[self.etat_0_canevas + i_etat_c], fill='turquoise3', 
                            outline = 'black')
        self.can.itemconfig(self.etats_texte_can[self.etat_0_canevas + i_etat_c], 
                                    text = self.machine.etats[i_etat_c], fill='black')

        dep_etat = (i_etat_s - i_etat_c) * self.largeur_cell_etat
        #print("dep etat : ", dep_etat, " etat courant ", etat_courant, " ", 
        #     self.machine.etats[etat_courant], " etat suivant ", etat_suivant, " ", 
        #     self.machine.etats[etat_suivant],"\n")
        if dep_etat > 0 :
            delta_dep = 1
        else:
            delta_dep = -1
        ratio = abs(dep_etat / self.largeur_cell_ruban)
        #print("ratio :", ratio, "dep_etat : ", dep_etat)
        cpt_ratio = 0         #  compteur pour le différentiel de vitesse d'animation
        cpt_px_tete = 0       #  compteur au cas où la tête n'arrive pas exactement centrée
        for i in range(abs(dep_etat)):
            self.x1_s = self.x1_s + delta_dep
            self.x0_s = self.x0_t
            self.can.delete(self.lien_etat)
            self.lien_etat = self.can.create_line(self.x0_s, self.y0_s, self.x1_s, self.y1_s, 
                                fill = 'OliveDrab4', width = 3)
            cpt_ratio = cpt_ratio + 1
            if cpt_ratio >= ratio :
                self.can.move(self.tete_lecture, dep, 0)
                cpt_ratio = 0
                self.x0_t = self.x0_t + dep
                
            self.can.update()
            self.can.after(self.vitesse)
        self.can.itemconfig(self.ruban_can[ind_ruban + dep], fill = 'firebrick3')
        self.can.itemconfig(self.ruban_texte_can[ind_ruban + dep], fill = 'white')
        self.can.itemconfig(self.etats_can[self.etat_0_canevas + i_etat_s], fill='turquoise4', 
                            outline = 'black')
        self.can.itemconfig(self.etats_texte_can[self.etat_0_canevas + i_etat_s], 
                                    text = self.machine.etats[i_etat_s], fill='white')
        self.can.update()
            
    def mvtEtatSeul(self, etat_courant, etat_suivant, ind_ruban, ecrit):
        #  uniquement pour la dernière étape de la machine
        liste_etats = list(self.machine.etats.keys())
        i_etat_c = liste_etats.index(etat_courant)
        i_etat_s = liste_etats.index(etat_suivant)
        self.can.itemconfig(self.etats_can[self.etat_0_canevas + i_etat_c], fill='turquoise3', 
                            outline = 'black')
        self.can.itemconfig(self.etats_texte_can[self.etat_0_canevas + i_etat_c], 
                                    text = self.machine.etats[i_etat_c], fill='black')

        dep_etat = (i_etat_s - i_etat_c) * self.largeur_cell_etat
        if dep_etat > 0 :
            delta_dep = 1
        else:
            delta_dep = -1
        for i in range(abs(dep_etat)):
            self.x1_s = self.x1_s + delta_dep
            self.x0_s = self.x0_t
            self.can.delete(self.lien_etat)
            self.lien_etat = self.can.create_line(self.x0_s, self.y0_s, self.x1_s, self.y1_s, 
                                fill = 'OliveDrab4', width = 3)
            self.can.update()
            self.can.after(self.vitesse)
            
        self.can.itemconfig(self.etats_can[self.etat_0_canevas + i_etat_s], fill='turquoise4', 
                            outline = 'black')
        self.can.itemconfig(self.etats_texte_can[self.etat_0_canevas + i_etat_s], 
                                    text = self.machine.etats[etat_suivant], fill='white')
        self.can.itemconfig(self.ruban_can[ind_ruban], fill = 'firebrick3')
        self.can.itemconfig(self.ruban_texte_can[ind_ruban], text = self.texte_symboles[ecrit], fill = 'white')
            
    def mouvement(self):
        # self.preparationMachine()
        #print("\nDans mouvement")
        #print(self.machine)
        self.deroulement = self.machine.deroulement

        
        for i in range(len(self.deroulement)):
            ruban_der = self.deroulement[i][0]
            ind_ruban = self.deroulement[i][1] + 2  #   décalage car 2 vides dans la gui
                                                    #  avant le ruban métier
            lu = self.deroulement[i][2]
            etat_courant = self.deroulement[i][3]
            ecrit = self.deroulement[i][4]
            etat_suivant = self.deroulement[i][5]
            dep = self.deroulement[i][6]
                        
            if etat_suivant >= self.machine.CODE_FIN:
                #  La tête ne se déplace pas sur un état final
                #   par contre le ruban peut changer
                #if dep is not None :    # à virer un jour
                self.mvtEtatSeul(etat_courant, etat_suivant, ind_ruban, ecrit)
                """
                else :
                    print("mise à jour du ruban uniquement lors de la dernière itération")
                    self.can.itemconfig(self.ruban_can[ind_ruban], fill = 'firebrick3')
                    #self.can.itemconfig(self.ruban_texte_can[ind_ruban], fill = 'white')
                    self.can.itemconfig(self.ruban_texte_can[ind_ruban], text = ecrit, fill = 'white')
               """
            elif etat_courant == etat_suivant:
                self.mvtTeteSeule(dep, ind_ruban, ecrit)
            else :
                self.mvtTeteEtat(dep, ind_ruban, ecrit, etat_courant, etat_suivant)
            if etat_suivant < self.machine.CODE_FIN :
                self.label_transition.configure(text= self.chaineTransition(i + 1))
                self.label_transition.update()
                self.can.after(1000)

