# -*- coding: utf-8 -*-
"""
Created on Mon Jun  7 08:21:56 2021

@author: Frédéric Mandon
www.maths-info-lycee.fr
"""
from random import randint

class MachineTuring():
    def __init__(self):
        # variables métier
        # les états finaux ont un code >= 10
        self.NB_TRANS = 100          #  nombre maximum de transitions effectuées
        self.ETAT_INIT = 'INIT'      #  nom de l'état d'initialisation
        self.CODE_INIT = 0           #  numéro de l'état initial. Ne pas modifier, plante la GUI
        self.ETAT_FIN = 'FIN'        #  nom de l'état final simple
        self.CODE_FIN = 9           #  code de l'état final. Ne pas modifier, plante la GUI
        self.ETAT_VRAI = 'VRAI'      #  nom de l'état Vrai 
        self.CODE_VRAI = 10          #  code de l'état final Vrai > 10. Ne pas modifier, plante la GUI
        self.ETAT_FAUX = 'FAUX'      #  nom de l'état Faux
        self.CODE_FAUX = 11          #  code de l'état final Faux > 10. Ne pas modifier, plante la GUI
        self.ETAT_NONE =  None       #  état inexistant
        self.CODE_NONE = -1          #  code des états inexistants
        self.LONGUEUR_RUBAN = 20     #  Longueur fixe du ruban
                                     #  obligatoire pour l'interface graphique
        self.NB_ETATS = 10           #  nombre maximum d'états (y compris INIT et FIN)
                                     #  il peut y en avoir moins
        self.deroulement = []        #  liste du déroulement 
                                     #  (ruban, indice position tête sur le ruban,
                                     #   symbole_lu, état courant, symbole écrit,
                                     #   état suivant, déplacement)
        self.etats = {self.CODE_INIT: self.ETAT_INIT}
        #  construction de la liste des états :
        #  0 : INIT, 1..9 : ETAT1..9 , 10 : FIN
        for i in range(1, self.NB_ETATS - 1):
            self.etats[i] = " ETAT " + str(i)
        self.etats[self.CODE_FIN] = self.ETAT_FIN  
        #  Contruction d'un ruban aléatoire
        #  Symboles 0, 1, 2 pour vide
        #  La 1ère case est vide
        self.ruban = [2]*self.LONGUEUR_RUBAN
        for i in range(1,self.LONGUEUR_RUBAN):
            valeur = randint(0,4)
            if valeur <= 3:
                self.ruban[i] = valeur // 2
        self.transitions = {}        #  transitions, dictionnaire 
                                     #  clés : couples (symbole lu, état courant)
                                     #  valeurs : triplet (symbole écrit, etat suivant
                                     #                     déplacement)
                                     #            déplacement +1 droite, -1 gauche, 0 rien
        self.resume = "Exemple de machine de Turing"

    def turing(self):
        """
        Génère la suite des rubans et des etats lors du fonctionnement
        d'une machine de Turing, maximum NB_TRANS transitions effectuées.
        La tête est positionnée à gauche (indice ruban 0), l'état initial est INIT
        @return ok : booléen Vrai si la machine de Turing n'a pas rencontré d'erreur
        """
        i = 0              #  compteur pour le nombre max de transition
        ind_ruban = 0
        while self.ruban[ind_ruban] == 2:
            ind_ruban = ind_ruban + 1
        ind_ruban = ind_ruban - 1      #  indice de la case lue sur le ruban
        etat_courant = self.CODE_INIT   #  code de l'état courant
        ok = True
        self.deroulement = []
        while ok and etat_courant < self.CODE_FIN and i < self.NB_TRANS :
            lu = self.ruban[ind_ruban]
            ok = False
            #  recherche de la bonne transition
            # print("lu : ", lu, " courant ",etat_courant)
            valeur = self.transitions.get((lu,etat_courant))
            if valeur is not None :
                (ecrit, etat_suivant, dep) = valeur
                ok = True
                self.deroulement.append([self.ruban[:], ind_ruban,
                                lu, etat_courant, ecrit, etat_suivant, dep])
            # print("avant :",self.ruban, " etat ", etat_courant, " indice ruban ", ind_ruban)
            etat_courant = etat_suivant
            self.ruban[ind_ruban] = ecrit
            ind_ruban = ind_ruban + dep
            # print("après :",self.ruban, " etat ", etat_courant," indice ruban ", ind_ruban,ok, "\n")
            i = i + 1
        """
        if self.ruban != self.deroulement[-1][0]:   # au cas où le ruban est modifié sur l'état final
            self.deroulement.append([self.ruban[:], ind_ruban,
                                None, etat_courant, None, etat_suivant, None])
        """
        return ok
    
    
    def exemple1(self):
        #  transitions, dictionnaire 
        #  clés : couples (symbole lu, état courant)
        #  valeurs : triplet (symbole écrit, etat suivant
        #                     déplacement)
        #            déplacement +1 droite, -1 gauche, 0 rien
        #  longueur ruban 20
        self.resume = "Inverse les 0 et les 1"
        self.etats = {self.CODE_INIT : self.ETAT_INIT , 
                      1 : "DROITE",
                      self.CODE_FIN : self.ETAT_FIN}
        self.transitions = {(2, 0):(2, self.CODE_INIT, 1),
                           (1, 0):(0, 1, 1),
                           (0, 0):(1, 1, 1),
                           (1, 1):(0, 1, 1),
                           (0, 1):(1, 1, 1),
                           (2, 1):(2, self.CODE_FIN, 0)
                           }
        self.ruban = [2] * 8 + [1, 1, 1, 0] + [2] * 8
        
    def exemple2(self):
        #  transitions, dictionnaire 
        #  clés : couples (symbole lu, état courant)
        #  valeurs : triplet (symbole écrit, etat suivant
        #                     déplacement)
        #            déplacement +1 droite, -1 gauche, 0 rien
        self.resume = "Des 0 puis des 1 ?"
        self.etats = {self.CODE_INIT : self.ETAT_INIT , 
                      1 : "DROITE",
                      2 : "TEST",
                      self.CODE_FAUX : self.ETAT_FAUX,
                      self.CODE_VRAI : self.ETAT_VRAI}
        self.transitions = {(2, 0):(2, self.CODE_INIT, 1),
                           (1, 0):(1, self.CODE_FAUX, 0),
                           (0, 0):(0, 1, 1),
                           (0, 1):(0, 1, 1),
                           (1, 1):(1, 2, 1),
                           (2, 1):(2, self.CODE_FAUX, 0),
                           (0, 2):(0, self.CODE_FAUX, 0),
                           (1, 2):(1, 2, 1),
                           (2, 2):(2, self.CODE_VRAI, 0)}
        
        self.ruban = [2] * 5 + [0, 0, 0, 0, 1, 1, 1, 1] + [2] * 7

    def exemple3(self):
        self.resume = "N fois 0 suivi de N fois 1 ?"
        self.etats = {self.CODE_INIT : self.ETAT_INIT , 
                      1 : "EFFACE_0",
                      2 : "DROITE_0",
                      3 : "DROITE_1",
                      4 : "PRE_EFF_1",
                      5 : "POST_EFF_1",
                      6 : "GAUCHE_1",
                      7 : "GAUCHE_0",
                      self.CODE_FAUX : self.ETAT_FAUX,
                      self.CODE_VRAI : self.ETAT_VRAI}
        self.transitions = {(2, 0):(2, self.CODE_INIT, 1),
                           (1, 0) :(1, self.CODE_FAUX, 0),
                           (0, 0) :(2, 1, 1),
        
                           (0, 1) :(0, 2, 1),
                           (1, 1) :(1, 3, 1),
                           (2, 1) :(0, self.CODE_FAUX, 0),
                           
                           (0, 2) :(0, 2, 1),
                           (1, 2) :(1, 3, 1),
                           (2, 2) :(2, self.CODE_FAUX, 0),
                           
                           (0, 3) : (0, self.CODE_FAUX, 0),
                           (1, 3) : (1, 3, 1),
                           (2, 3) : (2, 4, -1),
                           
                           (1, 4) : (2, 5, -1),
                           
                           (2, 5) : (2, self.CODE_VRAI, 0),
                           (1, 5) : (1, 6, -1),
                           (0, 5) : (0, 7, -1),
                           
                           (0, 6) : (0, 7, -1),
                           (1, 6) : (1, 6, -1),
                           (2, 6) : (2, self.CODE_FAUX, 0),
                           
                           (2, 7) : (2, self.CODE_INIT, 1),
                           (0, 7) : (0, 7, -1)}
        
        self.ruban = [2] * 6 + [0, 0, 0, 0, 1, 1, 1, 1] + [2] * 6 
        
    def exemple4(self):
        self.resume = "Calcul de n + 1 en binaire"
        self.etats = {self.CODE_INIT : self.ETAT_INIT , 
                      1 : "LECTURE",
                      2 : "CALCUL",
                      self.CODE_FIN : self.ETAT_FIN}
        self.transitions = {(2, 0):(2, self.CODE_INIT, 1),
                           (1, 0):(1, 1, 1),
                           (0, 0):(0, 1, 1),
                           (2, 1):(2, 2, -1),
                           (1, 1):(1, 1, 1),
                           (0, 1):(0, 1, 1),
                           (2, 2):(1, self.CODE_FIN, 0),
                           (1, 2):(0, 2, -1),
                           (0, 2):(1, self.CODE_FIN, 0)}
        
        self.ruban = [2] * 8 + [1, 1, 1, 1] + [2] * 8

    def __repr__(self):
        chaine = "Liste des états :\n"
        for  etat in self.etats:
            if etat != self.CODE_NONE :
                chaine = chaine + "état n° {} : {}\n".format(etat, self.etats[etat])
        chaine = chaine + "\nListe des transitions :\n"
        for transition in self.transitions :
            (lu,courant) =  transition
            (ecrit, suivant, dep) = self.transitions[transition]
            if lu == 2:
                lu = "vide"
            else :
                lu = str(lu)
            if ecrit == 2 :
                ecrit = "vide"
            else :
                ecrit = str(ecrit)
            courant = self.etats[courant]
            suivant = self.etats[suivant]
            if dep == 1:
                dep = "droite"
            elif dep == -1 :
                dep = "gauche"
            else :
                dep = "immobile"
            chaine_tr = "Symbole lu : {} , Etat courant : {} ,\tSymbole ecrit : {}, ".format(lu, courant, ecrit) +\
                "Etat suivant : {}, Déplacement : {}\n".format(suivant, dep)
            chaine = chaine + chaine_tr
        chaine = chaine + "Ruban :\n"
        for elt in self.ruban:
            chaine = chaine + str(elt)
        return chaine
        
def test():
    machine = MachineTuring()
    machine.exemple3()
    print(machine)
    ok = machine.turing()
    print(ok)
    if not ok:
        print("Vous foirâtes votre machine")
    for ligne in machine.deroulement :
        print(ligne)
    print("etat final : ",machine.etats[machine.deroulement[-1][5]])
